import sys
sys.path.append("/usr/local/lib/python3.6/dist-packages")

import math

import torch
import torch.nn as nn
import torch.nn.functional as f
from torch.distributions.normal import Normal
import numpy as np

import matplotlib.pyplot as plt
from pylab import rcParams

import torch.distributions as distributions

from pylab import rcParams
rcParams['figure.figsize'] = 10, 8
rcParams['figure.dpi'] = 300
rcParams['figure.figsize'] = 8, 8
plt.ion()


device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')


LSIZE = 32
ASIZE = 1
RSIZE = 256
#the coefficient used to scale initial u when u is not network output
u_coefficient = 0.4


class FRMDN(nn.Module):
    def __init__(self, K, block, layer, hidden):
        super(FRMDN, self).__init__()

        self.K = K
        self.block = block

        z_dim = 32
        self.z_dim = 32
        # h_dim = hidden
        h_dim = z_dim


        nets = lambda: nn.Sequential(nn.Linear(z_dim // 2, h_dim), nn.LeakyReLU(), nn.Linear(h_dim, z_dim // 2), nn.Tanh())
        nett = lambda: nn.Sequential(nn.Linear(z_dim // 2, h_dim), nn.LeakyReLU(), nn.Linear(h_dim, z_dim // 2))

        self.t = torch.nn.ModuleList([nett() for _ in range(self.block)])
        self.s = torch.nn.ModuleList([nets() for _ in range(self.block)])

    def f(self, x):
        log_det_J, z = x.new_zeros(x.shape[0], x.shape[1]), x
        for i in reversed(range(self.block)):
            if (i % 2 == 0):
                s = self.s[i](z[:, :, :self.z_dim // 2])
                t = self.t[i](z[:, :, :self.z_dim // 2])
                z1 = z[:, :, :self.z_dim // 2]
                z2 = (z[:, :, self.z_dim // 2:] - t) * torch.exp(-s)
                z = torch.cat((z1, z2), dim=2)
                log_det_J += s.sum(dim=2)
            else:
                s = self.s[i](z[:, :, self.z_dim // 2:])
                t = self.t[i](z[:, :, self.z_dim // 2:])
                z1 = z[:, :, self.z_dim // 2:]
                z2 = (z[:, :, self.z_dim // 2:] - t) * torch.exp(-s)
                z = torch.cat((z1, z2), dim=2)
                log_det_J += s.sum(dim=2)
        return z, log_det_J



def log_prob(z, mus, sigmas, logpi, U, logdet, is_train, Method,shared_u, d_prim_size):
    loss1 = gmm_loss(z.reshape(-1, 32, 32), mus, sigmas, logpi, U, reduce=True, is_train = is_train, shared_u = shared_u, d_prim = d_prim_size)
    if (Method != "rmdn"):
        loss2 = logdet
        loss = loss1 + loss2
    else:
        loss2 = torch.zeros(1).to(device)
        loss = loss1
    return dict(loss1=loss1.mean().mean(), loss2=loss2.mean().mean(), loss=loss.mean().mean())

def Determinant(is_train, U, D, d, d_prim):
    if (is_train):
        UtU = torch.det(torch.matmul(U.t(), U))
        UtU_root = torch.pow(UtU, (1/d_prim))
        det_D = torch.prod(D)
        det_D_root = torch.exp((-1 / d_prim) * torch.sum(D))
        term1 = det_D_root + UtU_root
        term2 = term1 * det_D
        term2_root = torch.pow(term2, (1/ d))
        term3 = d * torch.log(torch.abs(term2_root - 1))
        print("UtU {} UtU_root {} det_D_root {} term1 {}, term2 {} term2_root {}, abs {} term3 {}".format(UtU, UtU_root, det_D_root, term1, term2, term2_root, torch.abs(term2_root - 1), term3))
        return term3
    else:
        det = U @ torch.diag(D) @ U.t() + 0.1 * (torch.eye(32).to(device))
        return torch.log(torch.abs(torch.det(det)))

def gmm_loss(batch, mus, D, logpi, U, reduce=True, is_train =True ,shared_u=False, d_prim = 1): # pylint: disable=too-many-arguments

    # batch = batch.contiguous().view(-1, 32)
    # batch, _ = model_RNVP(batch)
    # batch = batch.contiguous().view(32, 16, 32)
    d = mus.size()[3]
    bsize = mus.size()[0]
    seq_len = mus.size()[1]
    guassians = mus.size()[2]
    batch = batch.unsqueeze(-2).to(device)
    y_mu = (batch - mus).to(device)
    if shared_u:
        #FIXME
        U = None
        # U = Us[seq_num,batch_num,:].repeat(num_of_guassians).view(num_of_guassians,d,d_prim)
    # else:
    #     U = Us[seq_num,:,:].view(batch_size,num_of_guassians,d,d_prim)
    uutranspose = torch.matmul(U, U.transpose(4,3)).to(device)
    inverse_covariance_det_term1 = torch.det(torch.eye(d_prim).to(device) + torch.matmul(torch.matmul(U.transpose(4,3),torch.diag_embed(torch.reciprocal(D))),U)).to(device)
    inverse_covariance_det_term2 = torch.prod(D,3).to(device)
    losses = logpi - \
           (d / 2) * math.log(2 * math.pi) + \
           0.5 * torch.log(inverse_covariance_det_term1) + 0.5 * torch.log(inverse_covariance_det_term2) - \
           0.5 * torch.matmul(torch.matmul(y_mu.unsqueeze(-1).transpose(4,3),torch.diag_embed(D)),y_mu.unsqueeze(-1)).view(bsize,seq_len,guassians) -\
           0.5 * torch.matmul(torch.matmul(y_mu.unsqueeze(-1).transpose(4,3),uutranspose),y_mu.unsqueeze(-1)).view(bsize,seq_len,guassians).to(device)


    print("-------------------->>>>>>>> min = {}, mean = {}, max = {}".format(torch.min(losses), torch.mean(losses), torch.max(losses)))

    log_prob = torch.logsumexp(losses,-1).to(device)
    del losses
    log_prob = torch.mean(log_prob,0)
    if reduce:
        return -torch.mean(log_prob)
    return -log_prob


class _MDRNNBase(nn.Module):
    def __init__(self, latents, actions, hiddens, gaussians, shared_U ,d_prim_size,u_is_network_output):
        super().__init__()
        self.latents = latents
        self.actions = actions
        self.hiddens = hiddens
        self.gaussians = gaussians

        if u_is_network_output:
            if shared_U:
                self.gmm_linear = nn.Linear(hiddens, ((2 * latents + 1) * gaussians) + (latents * d_prim_size))
            else:
                self.gmm_linear = nn.Linear(hiddens, (((2 + d_prim_size) * latents + 1) * gaussians))
        else:
            self.gmm_linear = nn.Linear(hiddens, ((2 * latents + 1) * gaussians))

    def forward(self, *inputs):
        pass


class MDRNN(_MDRNNBase):

    def __init__(self, latents, actions, hiddens, gaussians, shared_U, d_prim_size, u_is_network_output):
        super().__init__(latents, actions, hiddens, gaussians, shared_U , d_prim_size, u_is_network_output)
        self.rnn = nn.LSTM(latents + actions, hiddens)
        self.shared_u = shared_U
        self.u_is_net_output = u_is_network_output
        self.u_dimension = d_prim_size
        if u_is_network_output:
            self.U = None
        else:
            #FIXME: should U be the shape of [16,32,5,32,d_prim] or only [5,32,d_prim]
            if shared_U:
                self.U = torch.nn.Parameter(torch.rand(latents, d_prim_size), requires_grad=True)
            else:
                self.U = torch.nn.Parameter(
                    torch.from_numpy(np.random.rand(gaussians, latents, d_prim_size)),
                    requires_grad=True)


    def forward(self, actions, latents):

        seq_len, bs = actions.size(0), actions.size(1)
        ins = torch.cat([actions, latents], dim=-1)

        outs, _ = self.rnn(ins)
        gmm_outs = self.gmm_linear(outs)

        stride1 = self.gaussians * self.latents
        if (self.shared_u):
            stride2 = 2*stride1 + self.latents * self.u_dimension
        else:
            stride2 = 2*stride1 + self.latents * self.u_dimension * self.gaussians


        mus = gmm_outs[:, :, :stride1]
        mus = mus.view(seq_len, bs, self.gaussians, self.latents)

        if self.u_is_net_output:
            # ask: is there a reason to use if else on shared_u as well?is anything different?
            # ask: do I have to change shape of u with view or shape?if yes, is it
            #  different for shared and non shared U?
            # ask: can I assign gmm_output to self.u and don't use the variable u?(if it's ok to do this,
            #  remove variable u and return self.u

            sigmas = gmm_outs[:, :, stride1: 2*stride1]
            sigmas = sigmas.view(seq_len, bs, self.gaussians, self.latents)
            u = gmm_outs[:, :, 2*stride1:stride2]
            if self.shared_u:
                u = u.view(seq_len, bs,self.latents, self.u_dimension)
            else:
                u = u.view(seq_len, bs, self.gaussians, self.latents, self.u_dimension)
            pi = gmm_outs[:, :, stride2:]

        else:
            sigmas = gmm_outs[:, :, stride1: 2*stride1]
            sigmas = sigmas.view(seq_len, bs, self.gaussians, self.latents)
            u = self.U.repeat(seq_len,bs,1).view(seq_len, bs, self.gaussians, self.latents, self.u_dimension).to(dtype=torch.float32)
            pi = gmm_outs[:, :, 2 * stride1:]

        # if self.is_u == 'identity':
        #     sigmas = gmm_outs[:, :, stride1: 2*stride1]
        #     sigmas = sigmas.view(seq_len, bs, self.gaussians, self.latents)
        #     pi = gmm_outs[:, :, 2*stride1:]
        # else:
        #     sigmas = gmm_outs[:, :, stride1: stride1 + stride2]
        #     if (self.Type == "less"):
        #         sigmas = sigmas.view(seq_len, bs, self.gaussians, self.latents // 2)
        #     else:
        #         sigmas = sigmas.view(seq_len, bs, self.gaussians, self.latents * 2)
        #     pi = gmm_outs[:, :, stride1 + stride2:]


        #sigmas = torch.sqrt(torch.exp(sigmas)) + 0.01
        sigmas = torch.clamp(torch.sqrt(torch.exp(sigmas)), 0.01, 1.9)
        # sigmas = torch.sqrt(torch.exp(sigmas))
        # sigmas = torch.exp(sigmas)

        pi = pi.view(seq_len, bs, self.gaussians)
        logpi = f.log_softmax(pi, dim=-1)

        return mus, sigmas, logpi, u