import sys
sys.path.append('/usr/local/lib/python3.6/dist-packages')

import argparse
from functools import partial
from os.path import join, exists
from os import mkdir
import torch
import torch.nn.functional as f
from torch.utils.data import DataLoader
from torchvision import transforms
import numpy as np
from tqdm import tqdm
from utils.misc import save_checkpoint
from utils.misc import ASIZE, LSIZE, RSIZE, RED_SIZE, SIZE
from utils.learning import EarlyStopping
from utils.learning import ReduceLROnPlateau

from data.loaders import RolloutSequenceDataset
from models.vae import VAE
from models.mdrnn import FRMDN, MDRNN, log_prob

import csv
import os

parser = argparse.ArgumentParser("frmdn training")
parser.add_argument('--logdir', type=str, default='./logs',
                    help="Where things are logged and models are loaded from.")
parser.add_argument('--optim', type=str, default='rms',
                    help="optimizer")
parser.add_argument('--noreload', default=True, action='store_true',
                    help="Do not reload if specified.")
parser.add_argument('--include_reward', default=False, action='store_true',
                    help="Add a reward modelisation term to the loss.")
parser.add_argument('--K', type=int, default=5,
                    help="component")
parser.add_argument('--block', type=int, default=1,
                    help="block")
parser.add_argument('--epochs', type=int, default=20)
parser.add_argument('--lr', type=float, default=1e-5,
                    help="lr")
parser.add_argument('--layer', type=int, default=1,
                    help="layer")
parser.add_argument('--hidden', type=int, default=256,
                    help="hidden")
parser.add_argument('--M', type=str, default="rmdn",
                    help="method type")
#new added arguments
parser.add_argument('--U_is_network_output', type=bool, default=True,
                    help="U is output of network")
parser.add_argument('--d_prim_size', type=int, default=1,
                    help="dimension of d prim")
parser.add_argument('--shared_U', type=bool, default=False,
                    help="U is shared in components")
parser.add_argument('--complete_log_direction', type=str, default="guassians5/u_is_not_net_output/dprim1/",
                    help="U is shared in components")
args = parser.parse_args()
# removed
# parser.add_argument('--U', type=str, default="nonsquare", #="identity",
#                     help="U type")
# parser.add_argument('--T', type=str, default="less",
#                     help="method type")

# constants
BSIZE = 16
SEQ_LEN = 32
epochs = args.epochs
K = args.K
block = args.block
lr = args.lr
optim = args.optim
Method = args.M
#added
U_is_network_output = args.U_is_network_output
d_prim_size = args.d_prim_size
shared_U = args.shared_U
# removed
# is_u = args.U
# Type = args.T

label = 'mixture-{}-block-{}-epochs-{}-lr-{}-optim-{}-layer-{}-hidden-{}-shared_u-{}-dprim-{}-out-from-net-{}-{}-with-clamping-sigma'.format(args.K, args.block, args.epochs, args.lr, args.optim, args.layer, args.hidden,args.shared_U, args.d_prim_size, args.U_is_network_output , Method)

#server/local change needed
device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
# device = torch.device('cpu')
# Loading VAE
#server/local change needed
# vae_file = join('./logs/vae', 'best.tar')
vae_file = join('../../fatemeh/Original_WM/logs/vae', 'best.tar')
assert exists(vae_file), "No trained VAE in the logdir..."
#server/local change needed
# ,map_location=torch.device('cpu')
state = torch.load(vae_file)
print("Loading VAE at epoch {} "
      "with test error {}".format(
          state['epoch'], state['precision']))
#server/local change needed
# vae = VAE(3, LSIZE).to('cpu')
vae = VAE(3, LSIZE).to(device)
vae.load_state_dict(state['state_dict'])

# Loading model
# rnn_dir = join(args.logdir, 'frmdn_U/'+label)
rnn_dir = join(args.logdir, args.complete_log_direction +label)
rnn_file = join(rnn_dir, 'best.tar')
# print(rnn_dir)
if not exists(rnn_dir):
    mkdir(rnn_dir)


frmdn = FRMDN(K=K, block=block, layer=args.layer, hidden=args.hidden).to(device)
mdrnn = MDRNN(latents=LSIZE, actions=ASIZE, hiddens=args.hidden, gaussians=args.K,
              shared_U = shared_U,d_prim_size = d_prim_size,u_is_network_output = U_is_network_output).to(device)

# C = torch.nn.Parameter(torch.ones(1), requires_grad=True).to(device)
# loss1, loss2, loss = -frmdn.log_prob(x, action, latent_obs)

if optim == 'rms':
    if (Method != 'rmdn'):
        optimizer = torch.optim.RMSprop(list(frmdn.parameters()) + list(mdrnn.parameters()), lr=lr, alpha=.9)
    else:
        optimizer = torch.optim.RMSprop(list(mdrnn.parameters()), lr=lr, alpha=.9)
else:
    if (Method != 'rmdn'):
        optimizer = torch.optim.Adam(list(frmdn.parameters()) + list(mdrnn.parameters()), lr=lr)
    else:
        optimizer = torch.optim.Adam(list(mdrnn.parameters()), lr=lr)
scheduler = ReduceLROnPlateau(optimizer, 'min', factor=0.5, patience=5)
earlystopping = EarlyStopping('min', patience=30)


if exists(rnn_file) and not args.noreload:
    rnn_state = torch.load(rnn_file)
    print("Loading frmdn at epoch {} "
          "with test error {}".format(
              rnn_state["epoch"], rnn_state["precision"]))
    frmdn.load_state_dict(rnn_state["state_dict"])
    optimizer.load_state_dict(rnn_state["optimizer"])
    scheduler.load_state_dict(state['scheduler'])
    earlystopping.load_state_dict(state['earlystopping'])


# Data Loading
transform = transforms.Lambda(
    lambda x: np.transpose(x, (0, 3, 1, 2)) / 255)

#server/local change needed
#address: 'datasets/carracing'
train_loader = DataLoader(
    RolloutSequenceDataset('../../fatemeh/Original_WM/datasets/carracing', SEQ_LEN, transform, buffer_size=30),
    batch_size=BSIZE, num_workers=8, shuffle=True, pin_memory=False)
# valid_loader = DataLoader(
#     RolloutSequenceDataset('datasets/carracing', SEQ_LEN, transform, buffer_size=30),
#     batch_size=BSIZE, num_workers=8, shuffle=True)
test_loader = DataLoader(
    RolloutSequenceDataset('../../fatemeh/Original_WM/datasets/carracing', SEQ_LEN, transform, train=False, buffer_size=10),
    batch_size=BSIZE, num_workers=8)

def to_latent(obs, next_obs):
    with torch.no_grad():
        obs, next_obs = [
            f.upsample(x.view(-1, 3, SIZE, SIZE), size=RED_SIZE,
                       mode='bilinear', align_corners=True)
            for x in (obs, next_obs)]

        (obs_mu, obs_logsigma), (next_obs_mu, next_obs_logsigma) = [
            vae(x)[1:] for x in (obs, next_obs)]

        latent_obs, latent_next_obs = [
            (x_mu + x_logsigma.exp() * torch.randn_like(x_mu)).view(BSIZE, SEQ_LEN, LSIZE)
            for x_mu, x_logsigma in
            [(obs_mu, obs_logsigma), (next_obs_mu, next_obs_logsigma)]]
    return latent_obs, latent_next_obs

EpochCsv = open(os.path.join(args.logdir, args.complete_log_direction+Method+"-"+label+"-epoch.csv"), 'w', newline='')
# fieldnames_epoch = ['epoch', 'train', 'test']
# fieldnames_epoch = ['epoch', 'traing', 'trainp', 'trainl', 'testg', 'testp', 'testl', 'validg', 'validp', 'validl']
fieldnames_epoch = ['epoch', 'traing', 'trainp', 'trainl', 'testg', 'testp', 'testl']
writer_epoch = csv.DictWriter(EpochCsv, fieldnames=fieldnames_epoch)
writer_epoch.writeheader()

ParamsCsv = open(os.path.join(args.logdir, args.complete_log_direction +Method+"-"+label+"-params.csv"), 'w', newline='')
# fieldnames_params = ['epoch', 'trainS', 'trainM', 'trainA', 'testS', 'testM', 'testA']
fieldnames_params = ['epoch', 'trainSmin', 'trainSmean', 'trainSmax',
                                    'trainMmin', 'trainMmean', 'trainMmax',
                                    'trainAmin', 'trainAmean', 'trainAmax',
                     'testSmin', 'testSmean', 'testSmax',
                     'testMmin', 'testMmean', 'testMmax',
                     'testAmin', 'testAmean', 'testAmax']
writer_params = csv.DictWriter(ParamsCsv, fieldnames=fieldnames_params)
writer_params.writeheader()

def data_pass(epoch, train, include_reward, valid=False):
    if train:
        if Method != "rmdn":
            frmdn.train()
        mdrnn.train()
        loader = train_loader
    else:
        if Method != "rmdn":
            frmdn.eval()
        mdrnn.eval()
        loader = test_loader
        # if valid:
        #     loader = valid_loader
        # else:
        #     loader = test_loader

    loader.dataset.load_next_buffer()

    cum_loss = 0
    cum_gmm = 0
    cum_logp = 0
    # cum_bce = 0
    # cum_mse = 0

    S = {'min':[], 'mean':[], 'max':[]}
    M = {'min': [], 'mean': [], 'max': []}
    A = {'min': [], 'mean': [], 'max': []}

    for i, data in enumerate(loader):
        obs, action, reward, terminal, next_obs = [arr.to(device) for arr in data]

        latent_obs, latent_next_obs = to_latent(obs, next_obs)

        if train:

            if (Method != 'rmdn'):
                z, logdet = frmdn.f(latent_next_obs)
            else:
                z = latent_next_obs
                logdet = None

            mus, sigmas, logpi, U = mdrnn(action, latent_obs)
            losses = log_prob(z, mus, sigmas, logpi, U, logdet, train, Method, shared_U,d_prim_size)


            optimizer.zero_grad()
            print("\nsigmas ====== ", torch.min(sigmas).item(), torch.max(sigmas).item())
            print("mus ====== ", torch.min(mus).item(), torch.max(mus).item())
            print("logpi ====== ", torch.min(logpi).item(), torch.max(logpi).item(), "\n")



            losses['loss'].backward()
            #torch.nn.utils.clip_grad_norm_(mdrnn.parameters(), max_norm=2., norm_type=2)
            #torch.nn.utils.clip_grad_norm_(frmdn.parameters(), max_norm=1., norm_type=2)
            optimizer.step()
            # writer_iter_train.writerow({'epoch': e, 'iter': i, 'gmm':losses['loss1'].item(), 'logp':losses['loss2'].item(), 'loss': losses['loss'].item()})

#         elif not valid:
#             with torch.no_grad():
#                 # losses = get_loss(latent_obs, action, reward,
#                 #                   terminal, latent_next_obs, include_reward)
#                 # losses = log_prob(latent_next_obs, action, latent_obs)
#                 z, logdet = frmdn.f(latent_next_obs)
#
#                 # z, logp = self.f(x)
#                 # print('********************************')
#                 # print('action = ', action.shape)
#                 # print('latent obs = ', latent_obs.shape)
#                 mus, sigmas, logpi, U = mdrnn(action, latent_obs)
#
#                 losses = log_prob(z, mus, sigmas, logpi, U, logdet, 'valid')  # action, latent_obs)
# # log_prob(z, ):
#                 # writer_iter_test.writerow(
#                 #     {'epoch': e, 'iter': i, 'gmm': losses['loss1'].item(), 'logp': losses['loss2'].item(),
#                 #      'loss': losses['loss'].item()})

        else:
            with torch.no_grad():

                if (Method != 'rmdn'):
                    z, logdet = frmdn.f(latent_next_obs)
                else:
                    z = latent_next_obs
                    logdet = None

                mus, sigmas, logpi, U = mdrnn(action, latent_obs)

                losses = log_prob(z, mus, sigmas, logpi, U, logdet, train, Method, shared_U, d_prim_size)

        S['min'].append(torch.min(sigmas).item())
        S['mean'].append(torch.mean(sigmas).item())
        S['max'].append(torch.max(sigmas).item())

        M['min'].append(torch.min(mus).item())
        M['mean'].append(torch.mean(mus).item())
        M['max'].append(torch.max(mus).item())

        A['min'].append(torch.min(torch.exp(logpi)).item())
        A['mean'].append(torch.mean(torch.exp(logpi)).item())
        A['max'].append(torch.max(torch.exp(logpi)).item())

        # print("sigmas ====== ", torch.min(sigmas).item(), torch.max(sigmas).item())
        # print("mus ====== ", torch.min(mus).item(), torch.max(mus).item())
        # print("logpi ====== ", torch.min(logpi).item(), torch.max(logpi).item())

        cum_loss += losses['loss'].item()
        cum_gmm += losses['loss1'].item()
        cum_logp += losses['loss2'].item()

        print("loss={loss:10.6f} logp={logp:10.6f} gmm={gmm:10.6f}".format(loss=cum_loss / (i + 1), logp=cum_logp / (i + 1), gmm=cum_gmm/ (i + 1)))


    g = cum_gmm/ len(loader.dataset)
    p = cum_logp / len(loader.dataset)
    l = cum_loss / len(loader.dataset)

    return dict(g=g, p=p, l=l, S=S, M=M, A=A)

train = partial(data_pass, train=True, include_reward=args.include_reward)
test = partial(data_pass, train=False, include_reward=args.include_reward)
# valid = partial(data_pass, train=False, include_reward=args.include_reward, valid=True)


cur_best = None
for e in range(epochs):

    TestLoss = test(e)
    # ValidLoss = valid(e)
    TrainLoss = train(e)

    traing, trainp, trainl = TrainLoss['g'], TrainLoss['p'], TrainLoss['l']
    testg, testp, testl = TestLoss['g'], TestLoss['p'], TestLoss['l']
    # validg, validp, validl = ValidLoss['g'], ValidLoss['p'], ValidLoss['l']

    trainS, trainM, trainA = TrainLoss['S'], TrainLoss['M'], TrainLoss['A']
    testS, testM, testA = TestLoss['S'], TestLoss['M'], TestLoss['A']

    scheduler.step(testl)
    earlystopping.step(testl)

    writer_epoch.writerow({'epoch': e, 'traing': traing, 'trainp': trainp, 'trainl': trainl,
                                            'testg': testg, 'testp': testp, 'testl': testl})

    writer_params.writerow({'epoch':e,
                            'trainSmin':min(trainS['min']), 'trainSmean':np.mean(trainS['mean']), 'trainSmax':max(trainS['max']),
                             'trainMmin':min(trainM['min']), 'trainMmean':np.mean(trainM['mean']), 'trainMmax':max(trainM['max']),
                             'trainAmin':min(trainA['min']), 'trainAmean':np.mean(trainA['mean']), 'trainAmax':max(trainA['max']),
                             'testSmin':min(testS['min']), 'testSmean':np.mean(testS['mean']), 'testSmax':max(testS['max']),
                             'testMmin':min(testM['min']), 'testMmean':np.mean(testM['mean']), 'testMmax':max(testM['max']),
                             'testAmin':min(testA['min']), 'testAmean':np.mean(testA['mean']), 'testAmax':max(testA['max'])})

    is_best = not cur_best or testl < cur_best
    if is_best:
        cur_best = testl
    checkpoint_fname = join(rnn_dir, 'checkpoint.tar')
    save_checkpoint({
        "state_dict": frmdn.state_dict(),
        "optimizer": optimizer.state_dict(),
        'scheduler': scheduler.state_dict(),
        'earlystopping': earlystopping.state_dict(),
        "precision": testl,
        "epoch": e}, is_best, checkpoint_fname,
                    rnn_file)
    print("#########################END OF EPOCH {} ########################".format(e))
    if earlystopping.stop:
        print("End of Training because of early stopping at epoch {}".format(e))
        break