from scipy.io import wavfile
import os
import numpy as np
import pickle
#import librosa

class DataLoaderr(object):

    def __init__(self, hp, info, kind):
        self.name = hp.name
        self.info = info[self.name]
        self.audio_path = self.info['audio_path_{}'.format(kind)]
        self.pkl_path = self.info['pkl_path'].format(kind)
        self.hp = hp
        self.batch_size = 1 # self.batch_size = hp.batch_size
        self.create_iterator()

    def create_iterator(self):
        pkl = open(self.pkl_path, "rb")
        data = pickle.load(pkl)
        self.data_list = list(data.items())
        self.iterator = iter(self.data_list)

    def next_batch(self):
        audio = []
        names = []
        len = []
        for _ in range(self.batch_size):
            try:
                next_data = next(self.iterator)
            except:
                np.random.shuffle(self.data_list)
                self.iterator = iter(self.data_list)
                next_data = next(self.iterator)
            audio.append(self.read_waves(next_data[0]))
            names.append(next_data[0])
            len.append(next_data[1])
            # print(self.name, next_data[0], next_data[1])
        batch = {'audio': audio, 'names': names, 'len': len}

        return self.padding(batch)

    def read_waves(self, name):
        file_path = os.path.join(self.audio_path, name)
        samplerate, data = wavfile.read(file_path)
        # data, samplerate = librosa.load(file_path)
        data = data[:len(data) - (len(data) % self.hp.x_dim)]
        data = [float(i) for i in data]
        # d =  (np.array(list(data)) - self.info['mu']) / self.info['var']
        if self.hp.make_noise:
            return np.round(np.random.normal(0, 0.1, len(list(data))) +  (np.array(list(data)) - self.info['mu']) / self.info['var'], 2)
        else:
            return np.round((np.array(list(data)) - self.info['mu']) / self.info['var'], 2)
        # return np.array(list(data))
        # return np.round(np.random.normal(0, 0.1, len(list(data))) +  (np.array(list(data)) - self.info['mu']) / self.info['var'], 2)

    def padding(self, batch):
        # batch = {'audio': audio, 'names': names, 'len': len}
        max_len = max([i.shape[0] for i in batch['audio']])
        # audio = []
        for i in range(len(batch['len'])):
            audio_i = np.append(batch['audio'][i], np.zeros(max_len - len(batch['audio'][i])))
            # audio.append(audio_i[np.newaxis, :].reshape((-1, hp.x_dim)))
            #     arrayToAppend = totalArray = np.zeros((4, 200))
            #     alist.append(arrayToAppend)
            # arr = np.concatenate(alist, axis=1)
            if (i == 0):
                audio = audio_i.copy()
                audio = audio.reshape((-1, self.hp.x_dim))[np.newaxis, :, :]
                # audio = audio.reshape((-1, self.hp.x_dim))
                # audio = audio.view(-1, self.hp.x_dim)[np.newaxis, :, :]
            else:
                # audio = np.concatenate(([audio, audio_i.reshape((-1, hp.x_dim))[np.newaxis, :, :]]), axis=0)
                audio = np.concatenate([audio, audio_i.reshape((-1, self.hp.x_dim))[np.newaxis, :, :]])
                # audio = np.concatenate([audio, audio_i.view(-1, self.hp.x_dim)[np.newaxis, :, :]])
                # audio = np.stack((audio, audio_i.reshape((-1, hp.x_dim))), axis=0)
        # np.stack(audio,ax)
        # pad_batch = {'audio': torch.from_numpy(audio).type(torch.FloatTensor), 'names': batch['names'], 'len': batch['len']}
        pad_batch = {'audio': audio, 'names': batch['names'], 'len': batch['len']}
        return pad_batch